from __future__ import annotations

from typing import Any

import numpy as np

from sasktran2.atmosphere import Atmosphere
from sasktran2.optical.base import OpticalProperty
from sasktran2.util.interpolation import linear_interpolating_matrix

from .base import Constituent


class NumberDensityScatterer(Constituent):
    def __init__(
        self,
        optical_property: OpticalProperty,
        altitudes_m: np.array,
        number_density: np.array,
        out_of_bounds_mode: str = "zero",
        **kwargs,
    ) -> None:
        """
        A scattering constituent that is defined by a number density on an altitude grid and an optical property

        Parameters
        ----------
        optical_property : OpticalProperty
            The optical property defining the scattering information
        altitudes_m : np.array
            The altitude grid in [m]
        number_density : np.array
            Number density in [m^-3]
        out_of_bounds_mode : str, optional
            Interpolation mode outside of the boundaries, "extend" and "zero" are supported, by default "zero"
        kwargs : dict
            Additional arguments to pass to the optical property.
        """
        super().__init__()

        self._out_of_bounds_mode = out_of_bounds_mode
        self._altitudes_m = altitudes_m
        self._number_density = number_density
        self._optical_property = optical_property

        # Extra factor to apply to the vertical derivatives, used by the derived Extinction class
        self._vertical_deriv_factor = np.ones_like(number_density)

        # Optical derivatives can also have derivatives to this factor
        self._d_vertical_deriv_factor = {}

        self._wf_name = "number_density"

        self._kwargs = kwargs

    def __getattr__(self, __name: str) -> Any:
        if __name in self.__dict__.get("_kwargs", {}):
            return self._kwargs[__name]
        return None

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name in self.__dict__.get("_kwargs", {}):
            self._kwargs[__name] = __value
        else:
            super().__setattr__(__name, __value)

    @property
    def number_density(self):
        return self._number_density

    @number_density.setter
    def number_density(self, number_density: np.array):
        self._number_density = number_density

    def add_to_atmosphere(self, atmo: Atmosphere):
        interp_matrix = linear_interpolating_matrix(
            self._altitudes_m,
            atmo.model_geometry.altitudes(),
            self._out_of_bounds_mode.lower(),
        )

        interped_kwargs = {k: interp_matrix @ v for k, v in self._kwargs.items()}

        self._optical_quants = self._optical_property.atmosphere_quantities(
            atmo, **interped_kwargs
        )

        interp_numden = interp_matrix @ self._number_density

        atmo.storage.total_extinction += (
            self._optical_quants.extinction * (interp_numden)[:, np.newaxis]
        )

        # Optical quants in SSA temporarily stores the SSA * extinction
        atmo.storage.ssa += self._optical_quants.ssa * (interp_numden)[:, np.newaxis]

        atmo.storage.leg_coeff += (
            self._optical_quants.ssa[np.newaxis, :, :]
            * (interp_numden)[np.newaxis, :, np.newaxis]
            * self._optical_quants.leg_coeff
        )

        # Convert back to SSA for ease of use later in the derivatives
        self._optical_quants.ssa /= self._optical_quants.extinction
        self._optical_quants.ssa[~np.isfinite(self._optical_quants.ssa)] = 1

    def register_derivative(self, atmo: Atmosphere, name: str):
        interp_matrix = linear_interpolating_matrix(
            self._altitudes_m,
            atmo.model_geometry.altitudes(),
            self._out_of_bounds_mode.lower(),
        )
        interped_kwargs = {k: interp_matrix @ v for k, v in self._kwargs.items()}

        derivs = {}

        # Factor to apply to legendre derivatives is
        # (species_ext) * (species_ssa) / (total_ext * total_ssa)

        deriv_mapping = atmo.storage.get_derivative_mapping(
            f"wf_{name}_{self._wf_name}"
        )

        deriv_mapping.d_extinction[:] += self._optical_quants.extinction
        deriv_mapping.d_ssa[:] += (
            self._optical_quants.extinction
            * (self._optical_quants.ssa - atmo.storage.ssa)
            / atmo.storage.total_extinction
        )
        deriv_mapping.d_leg_coeff[:] += (
            self._optical_quants.leg_coeff - atmo.storage.leg_coeff
        )
        deriv_mapping.scat_factor[:] += (
            self._optical_quants.ssa * self._optical_quants.extinction
        ) / (atmo.storage.ssa * atmo.storage.total_extinction)
        deriv_mapping.interpolator = (
            interp_matrix * self._vertical_deriv_factor[np.newaxis, :]
        )
        deriv_mapping.interp_dim = f"{name}_altitude"

        optical_derivs = self._optical_property.optical_derivatives(
            atmo, **interped_kwargs
        )

        for key, val in optical_derivs.items():
            deriv_mapping = atmo.storage.get_derivative_mapping(f"wf_{name}_{key}")

            deriv_mapping.d_extinction[:] += val.d_extinction
            # First, the optical property returns back d_scattering extinction in the d_ssa container,
            # convert this to d_ssa
            deriv_mapping.d_ssa[:] += (
                val.d_ssa - val.d_extinction * self._optical_quants.ssa
            ) / self._optical_quants.extinction
            deriv_mapping.d_leg_coeff[:] += val.d_leg_coeff

            if key in self._d_vertical_deriv_factor:
                # Have to make some adjustments

                # The change in extinction is adjusted
                deriv_mapping.d_extinction[:] += (
                    self._optical_quants.extinction
                    / (interp_matrix @ self._vertical_deriv_factor)[:, np.newaxis]
                    * (interp_matrix @ self._d_vertical_deriv_factor[key])[
                        :, np.newaxis
                    ]
                )

                # Change in single scatter albedo should be invariant whether or not we are
                # in extinction space or number density space

            # Start with leg_coeff
            deriv_mapping.d_leg_coeff[:] += (
                self._optical_quants.leg_coeff - atmo.storage.leg_coeff
            ) * (
                1 / self._optical_quants.ssa * deriv_mapping.d_ssa
                + 1 / self._optical_quants.extinction * deriv_mapping.d_extinction
            )[
                np.newaxis, :, :
            ]

            # Then adjust d_ssa
            deriv_mapping.d_ssa[:] *= self._optical_quants.extinction
            deriv_mapping.d_ssa[:] += deriv_mapping.d_extinction * (
                self._optical_quants.ssa - atmo.storage.ssa
            )
            deriv_mapping.d_ssa[:] /= atmo.storage.total_extinction

            # TODO: The model should probably handle this
            norm_factor = deriv_mapping.d_leg_coeff.max(axis=0)
            norm_factor[norm_factor == 0] = 1

            deriv_mapping.scat_factor[:] = (
                self._optical_quants.ssa * self._optical_quants.extinction
            ) / (atmo.storage.ssa * atmo.storage.total_extinction)

            deriv_mapping.d_leg_coeff[:] /= norm_factor[np.newaxis, :, :]
            deriv_mapping.scat_factor[:] *= norm_factor

            deriv_mapping.interpolator = (
                interp_matrix * self._number_density[np.newaxis, :]
            )
            deriv_mapping.interp_dim = f"{name}_altitude"

        return derivs


class ExtinctionScatterer(NumberDensityScatterer):
    def __init__(
        self,
        optical_property: OpticalProperty,
        altitudes_m: np.array,
        extinction_per_m: np.array,
        extinction_wavelength_nm: float,
        out_of_bounds_mode: str = "zero",
        **kwargs,
    ) -> None:
        """
        A scattering constituent that is defined by a number density on an altitude grid and an optical property

        Parameters
        ----------
        optical_property : OpticalProperty
            The optical property defining the scattering information
        altitudes_m : np.array
            The altitude grid in [m]
        extinction_per_m : np.array
            Extinction in [m^-1]
        extinction_wavelength_nm : float
            Wavelength that the extinction profile is specified at
        out_of_bounds_mode : str, optional
            Interpolation mode outside of the boundaries, "extend" and "zero" are supported, by default "zero"
        kwargs : dict
            Additional arguments passed to the optical property
        """
        self._extinction_per_m = extinction_per_m
        self._extinction_wavelength_nm = extinction_wavelength_nm

        super().__init__(
            optical_property, altitudes_m, None, out_of_bounds_mode, **kwargs
        )
        self._extinction_to_numden_factors = None
        self._update_numberdensity()
        self._wf_name = "extinction"

    def _update_numberdensity(self):
        self._extinction_to_numden_factors = self._optical_property.cross_sections(
            np.array([self._extinction_wavelength_nm]),
            altitudes_m=self._altitudes_m,
            **self._kwargs,
        ).extinction.flatten()
        self._vertical_deriv_factor = 1 / self._extinction_to_numden_factors

        self.number_density = (
            self._extinction_per_m / self._extinction_to_numden_factors
        )

        self._d_vertical_deriv_factor = (
            self._optical_property.cross_section_derivatives(
                np.array([self._extinction_wavelength_nm]),
                altitudes_m=self._altitudes_m,
                **self._kwargs,
            )
        )

        for _, val in self._d_vertical_deriv_factor.items():
            # convert from derivative of x to derivative of 1/x
            val *= -1 * self._vertical_deriv_factor**2  # noqa: PLW2901

    @property
    def extinction_per_m(self):
        return self._extinction_per_m

    @extinction_per_m.setter
    def extinction_per_m(self, extinction: np.array):
        self._extinction_per_m = extinction

    def add_to_atmosphere(self, atmo: Atmosphere):
        self._update_numberdensity()
        super().add_to_atmosphere(atmo)



from __future__ import annotations

import numpy as np


def celsius_to_kelvin(temperature_c: np.array):
    return temperature_c + 273.15


def kelvin_to_celsius(temperature_k: np.array):
    return temperature_k - 273.15


def wavlength_nm_to_wavenumber_cminv(wavelengths_nm: np.array):
    return 1e7 / wavelengths_nm


def wavenumber_cminv_to_wavlength_nm(wavenumber_cminv: np.array):
    return 1e7 / wavenumber_cminv



from __future__ import annotations

import numpy as np
import sasktran2 as sk

from . import parameterized

nwav = 1000


@parameterized(["nlyr", "calc_deriv"], ([2, 20, 100], [False, True]))
class TwoStreamNadirPlaneParallel:
    def setup(self, nlyr, calc_deriv):
        cos_sza = 0.5

        config = sk.Config()
        config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
        config.single_scatter_source = sk.SingleScatterSource.DiscreteOrdinates

        config.num_streams = 2
        config.num_singlescatter_moments = 4
        config.num_stokes = 1

        model_geometry = sk.Geometry1D(
            cos_sza=cos_sza,
            solar_azimuth=0,
            earth_radius_m=6372000,
            altitude_grid_m=np.linspace(0, 100000, nlyr + 1),
            interpolation_method=sk.InterpolationMethod.LinearInterpolation,
            geometry_type=sk.GeometryType.PlaneParallel,
        )

        viewing_geo = sk.ViewingGeometry()

        viewing_geo.add_ray(sk.GroundViewingSolar(cos_sza, np.deg2rad(30), 0.02, 2.0))
        viewing_geo.add_ray(sk.GroundViewingSolar(cos_sza, np.deg2rad(60), 0.92, 2.0))

        self._atmosphere = sk.Atmosphere(
            model_geometry, config, calculate_derivatives=calc_deriv, numwavel=nwav
        )

        self._atmosphere.storage.total_extinction[:] = 0.5 / 1.0
        self._atmosphere.storage.ssa[:] = 1

        self._atmosphere.leg_coeff.a1[0, :, :] = 1
        self._atmosphere.leg_coeff.a1[2, :, :] = 0.5

        self._atmosphere.surface.albedo[:] = 0.0

        self._engine = sk.Engine(config, model_geometry, viewing_geo)

    def time_two_stream_nadir(self, nlyr, calc_deriv):  # noqa: ARG002
        self._engine.calculate_radiance(self._atmosphere)



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def test_vandaele():
    """
    Tests the NO2Vandaele optical property against a few hardcoded test results obtained from SASKTRAN legacy
    """
    vand = sk.optical.NO2Vandaele()

    config = sk.Config()

    altitude_grid = np.arange(0, 65001, 1000)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    # Old SASKTRAN test results
    test_cases = []
    test_cases.append(
        {
            "temperature": 240,
            "values": np.array(
                [1.82452103e-19, 3.14023410e-19, 3.91827004e-19, 4.61405350e-20]
            ),
        }
    )
    test_cases.append(
        {
            "temperature": 228,
            "values": np.array(
                [1.81900948e-19, 3.11019054e-19, 3.85870165e-19, 4.60555094e-20]
            ),
        }
    )

    # once again old sasktran is in air wavelengths
    wavel = sk.optical.air_wavelength_to_vacuum_wavelength(
        np.array([310, 330, 350, 600])
    )

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    for test_case in test_cases:
        atmosphere.temperature_k[:] = test_case["temperature"]

        result = vand.atmosphere_quantities(atmo=atmosphere).extinction[10, :]

        np.testing.assert_allclose(result, test_case["values"] * 1e-4, rtol=1e-5)


def test_no2_hitran_compare():
    vand = sk.optical.NO2Vandaele()
    hitran = sk.optical.HITRANUV("no2")

    config = sk.Config()

    altitude_grid = np.arange(0, 65001, 1000)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    wavel = np.arange(350, 500, 0.001)

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere.temperature_k[:] = 240

    xs_vand = vand.atmosphere_quantities(atmo=atmosphere).extinction[10, :]
    xs_hitran = hitran.atmosphere_quantities(atmo=atmosphere).extinction[10, :]

    np.testing.assert_allclose(xs_hitran, xs_vand, rtol=1e-3)



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def _test_scenarios():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    altitude_grid = np.arange(0, 65001, 1000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geo.add_ray(sk.TangentAltitudeSolar(tan_alt, 0, 600000, 0.6))

    wavel = np.array([310, 330, 350, 600])

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    vmr_altitude_grid = np.array([10000.0, 30000.0, 60000.0])
    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(), vmr_altitude_grid, np.ones_like(vmr_altitude_grid) * 1e-6
    )
    atmosphere.surface.albedo[:] = 0.3

    atmos = []
    atmos.append(atmosphere)

    scen = []

    for atmo in atmos:
        scen.append(
            {
                "config": config,
                "geometry": geometry,
                "viewing_geo": viewing_geo,
                "atmosphere": atmo,
            }
        )

    return scen


def test_scattering_db_construction():
    mie = sk.optical.database.OpticalDatabaseGenericScatterer(
        sk.appconfig.database_root().joinpath("cross_sections/mie/sulfate_test.nc")
    )

    config = sk.Config()

    altitude_grid = np.arange(0, 65001, 1000)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    wavel = np.arange(350, 500, 0.1)

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    _ = mie.atmosphere_quantities(
        atmo=atmosphere,
        lognormal_median_radius=np.ones_like(atmosphere.model_geometry.altitudes())
        * 105,
    ).extinction[10, :]


def test_scattering_db_wf():
    """
    Tests that weighting functions with respect to the scattering DB auxillary variables (e.g. particle size)
    is working correctly

    The precision of these weighting functions can be quite poor
    """
    scens = _test_scenarios()

    for scen in scens:
        atmosphere = scen["atmosphere"]

        atmosphere["strat_aerosol"] = sk.test_util.scenarios.test_aerosol_constituent(
            atmosphere.model_geometry.altitudes(), extinction_space=False
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        atmosphere["strat_aerosol"].lognormal_median_radius[:] = 107

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere["strat_aerosol"].lognormal_median_radius,
            0.0001,
            engine,
            atmosphere,
            "wf_strat_aerosol_lognormal_median_radius",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_strat_aerosol_lognormal_median_radius"],
            radiance["wf_strat_aerosol_lognormal_median_radius_numeric"],
            wf_dim="strat_aerosol_altitude",
            decimal=3,
        )


def test_scattering_db_wf_extinction():
    """
    Tests that weighting functions with respect to the scattering DB auxillary variables (e.g. particle size)
    is working correctly

    The precision of these weighting functions can be quite poor
    """
    scens = _test_scenarios()

    for scen in scens:
        atmosphere = scen["atmosphere"]

        atmosphere["strat_aerosol"] = sk.test_util.scenarios.test_aerosol_constituent(
            atmosphere.model_geometry.altitudes(), extinction_space=True
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        atmosphere["strat_aerosol"].lognormal_median_radius[:] = 107

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere["strat_aerosol"].lognormal_median_radius,
            0.000001,
            engine,
            atmosphere,
            "wf_strat_aerosol_lognormal_median_radius",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_strat_aerosol_lognormal_median_radius"],
            radiance["wf_strat_aerosol_lognormal_median_radius_numeric"],
            wf_dim="strat_aerosol_altitude",
            decimal=2,
        )



from __future__ import annotations

from sasktran2._core import Geodetic


class WGS84(Geodetic):
    def __init__(self):
        """
        A geodetic object based upon the standard WGS84 ellipsoid. See
        :py:class:`sasktran2.Geodetic` for more information and usage.
        """
        super().__init__(6378137.0, 1.0 / 298.257223563)

    def __repr__(self):
        if self.valid:
            return f"WGS84 Location:\nLatitude: {self.latitude}, Longitude: {self.longitude}, Altitude: {self.altitude}"
        else:  # noqa: RET505
            return "WGS84 Unitialized"


class SphericalGeoid(Geodetic):
    def __init__(self, radius: float):
        """
        A geoid that is represented as a perfect sphere. See
        :py:class:`sasktran2.Geodetic` for more information and usage.

        Parameters
        ----------
        radius: float
            Radius of the sphere in [m]
        """
        super().__init__(radius, 0.0)

    def __repr__(self):
        if self.valid:
            return f"Spherical Geoid Location:\nLatitude: {self.latitude}, Longitude: {self.longitude}, Altitude: {self.altitude}"
        else:  # noqa: RET505
            return "Spherical Geoid Unitialized"



from __future__ import annotations

import numpy as np
import pandas as pd
import sasktran2 as sk


def test_limb_viewing_ray_conversion():
    time = pd.Timestamp.now()

    solar = sk.solar.SolarGeometryHandlerForced(45, 180)

    tan_alts = np.array([10, 20, 30, 40]) * 1000

    obs_geo = sk.WGS84()
    obs_geo.from_lat_lon_alt(20, 30, 700000)

    tp_geo = sk.WGS84()

    for tan_alt in tan_alts:
        lv = tp_geo.from_tangent_altitude(tan_alt, obs_geo.location, [0, 0, 1])

        ray = sk.viewinggeo.ecef.ecef_to_sasktran2_ray(
            obs_geo.location, lv, time, sk.WGS84(), solar
        )

        assert isinstance(ray, sk.TangentAltitudeSolar)

        string = str(ray).split(": ", 1)[1]

        # Split into individual parameter pairs
        pairs = string.split(", ")

        # Convert to dictionary
        param_dict = {}
        for pair in pairs:
            key, value = pair.split(": ")
            param_dict[key.strip()] = float(value)

        np.testing.assert_allclose(param_dict["cos_sza"], np.cos(np.deg2rad(45)))
        np.testing.assert_allclose(
            param_dict["relative_azimuth_angle"], np.deg2rad(180)
        )
        np.testing.assert_allclose(
            param_dict["tangentaltitude"], tan_alt, atol=0.1
        )  # Check to accurate within 0.1 m
        np.testing.assert_allclose(param_dict["observeraltitude"], 700000)



from __future__ import annotations

import sasktran2 as sk
from sasktran2.viewinggeo.base import ViewingGeometryContainer


class Engine:
    def __init__(
        self,
        config: sk.Config,
        model_geometry: sk.Geometry1D,
        viewing_geo: sk.ViewingGeometry,
    ):
        """
        An Engine is the main class that handles the radiative transfer calculation.  The calculation takes
        place in two components.

        First, upon construction of the Engine, the majority of the geometry information is computed and
        cached.

        The main calculation takes place when calling :py:meth:`~calculate_radiance` with an
        :py:class:`sasktran2.Atmosphere` object where the actual radiative transfer calculation
        is performed.

        Parameters
        ----------
        config : sk.Config
            Configuration object
        model_geometry : sk.Geometry1D
            Geometry for the model
        viewing_geo : sk.ViewingGeometry
            Viewing geometry
        """
        self._nstokes = config.num_stokes

        if self._nstokes == 1:
            self._engine = sk.EngineStokes_1(config, model_geometry, viewing_geo)
        elif self._nstokes == 3:
            self._engine = sk.EngineStokes_3(config, model_geometry, viewing_geo)

        self._model_geometry = model_geometry
        self._viewing_geometry = viewing_geo

    def calculate_radiance(self, atmosphere: sk.Atmosphere, output: sk.Output = None):
        """
        Performs the radiative transfer calculation for a given atmosphere

        Parameters
        ----------
        atmosphere : sk.Atmosphere
            _description_
        output : sk.Output, optional
            Optional abstract output type.  Can be specified to change the
            output format. If set to None, the default :py:class:`sasktran2.OutputIdeal` class
            is used, by default None

        Returns
        -------
        varies
            Exact return type depends upon what output is set to.  For the default
            :py:class:`sasktran2.OutputIdeal`, the output format is an `xarray` Dataset.
            See for example, :py:meth:`sasktran2.OutputIdeal.post_process`
        """
        engine_output = (
            sk.OutputDerivMapped(self._nstokes) if output is None else output
        )

        self._engine.calculate_radiance(
            atmosphere.internal_object(), engine_output.internal_output()
        )

        result = engine_output.post_process(
            atmosphere, self._model_geometry, self._viewing_geometry
        )

        if isinstance(self._viewing_geometry, ViewingGeometryContainer):
            result = self._viewing_geometry.add_geometry_to_radiance(result)

        return result



from __future__ import annotations

import numpy as np
import sasktran2 as sk

from . import _skip_slow, parameterized

nwav = 1000
nlyr = 100
nlos = 50


@parameterized(["nmoments", "calc_deriv"], ([4, 16], [False, True]))
class LimbSingleScatter:
    def setup(self, nmoments, calc_deriv):
        if nmoments == 16 and calc_deriv:
            _skip_slow()

        cos_sza = 0.5

        config = sk.Config()
        config.multiple_scatter_source = sk.MultipleScatterSource.NoSource

        config.num_singlescatter_moments = nmoments
        config.num_stokes = 1

        model_geometry = sk.Geometry1D(
            cos_sza=cos_sza,
            solar_azimuth=0,
            earth_radius_m=6372000,
            altitude_grid_m=np.linspace(0, 100000, nlyr + 1),
            interpolation_method=sk.InterpolationMethod.LinearInterpolation,
            geometry_type=sk.GeometryType.Spherical,
        )

        viewing_geo = sk.ViewingGeometry()

        for i in range(nlos):
            viewing_geo.add_ray(
                sk.TangentAltitudeSolar(i * (100000 / (nlos + 1)), 0, 200000, cos_sza)
            )

        self._atmosphere = sk.Atmosphere(
            model_geometry, config, calculate_derivatives=calc_deriv, numwavel=nwav
        )

        self._atmosphere.storage.total_extinction[:] = 0.5 / 1.0
        self._atmosphere.storage.ssa[:] = 1

        self._atmosphere.leg_coeff.a1[0, :, :] = 1
        self._atmosphere.leg_coeff.a1[2, :, :] = 0.5

        self._atmosphere.surface.albedo[:] = 0.0

        self._engine = sk.Engine(config, model_geometry, viewing_geo)

    def time_limb_single_scatter(self, nmoments, calc_deriv):  # noqa: ARG002
        self._engine.calculate_radiance(self._atmosphere)



# ruff: noqa: F401
from __future__ import annotations

from . import (
    appconfig,
    climatology,
    constituent,
    database,
    mie,
    optical,
    solar,
    spectroscopy,
    test_util,
    util,
    viewinggeo,
)
from ._core import (
    AtmosphereStokes_1,
    AtmosphereStokes_3,
    AtmosphereStorageStokes_1,
    AtmosphereStorageStokes_3,
    Config,
    EmissionSource,
    EngineStokes_1,
    EngineStokes_3,
    Geodetic,
    Geometry1D,
    GeometryType,
    GroundViewingSolar,
    InputValidationMode,
    InterpolationMethod,
    MultipleScatterSource,
    OccultationSource,
    OutputDerivMappedStokes_1,
    OutputDerivMappedStokes_3,
    OutputIdealStokes_1,
    OutputIdealStokes_3,
    SingleScatterSource,
    StokesBasis,
    SurfaceStokes_1,
    SurfaceStokes_3,
    TangentAltitudeSolar,
    ThreadingModel,
    ViewingGeometry,
    ViewingGeometryBase,
)
from ._version import __version__
from .atmosphere import Atmosphere
from .engine import Engine
from .geodetic import WGS84, SphericalGeoid
from .output import Output, OutputDerivMapped, OutputIdeal



from __future__ import annotations

import numpy as np
import sasktran2 as sk

from . import _skip_slow, parameterized

nwav = 1


@parameterized(
    ["nlyr", "calc_deriv", "nstr", "nstokes"],
    ([2, 20, 100], [False], [2, 4, 8, 16, 32], [1, 3]),
)
class DOLarge:
    def setup(self, nlyr, calc_deriv, nstr, nstokes):
        if nstr > 8 and calc_deriv:
            _skip_slow()

        cos_sza = 0.5

        config = sk.Config()
        config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
        config.single_scatter_source = sk.SingleScatterSource.DiscreteOrdinates

        config.num_streams = nstr
        config.num_singlescatter_moments = nstr + 1
        config.num_stokes = nstokes

        model_geometry = sk.Geometry1D(
            cos_sza=cos_sza,
            solar_azimuth=0,
            earth_radius_m=6372000,
            altitude_grid_m=np.linspace(0, 100000, nlyr + 1),
            interpolation_method=sk.InterpolationMethod.LinearInterpolation,
            geometry_type=sk.GeometryType.PlaneParallel,
        )

        viewing_geo = sk.ViewingGeometry()

        viewing_geo.add_ray(sk.GroundViewingSolar(cos_sza, np.deg2rad(30), 0.02, 2.0))
        viewing_geo.add_ray(sk.GroundViewingSolar(cos_sza, np.deg2rad(60), 0.92, 2.0))

        self._atmosphere = sk.Atmosphere(
            model_geometry, config, calculate_derivatives=calc_deriv, numwavel=nwav
        )

        self._atmosphere.storage.total_extinction[:] = 0.5 / 1.0
        self._atmosphere.storage.ssa[:] = 1

        self._atmosphere.leg_coeff.a1[0, :, :] = 1
        self._atmosphere.leg_coeff.a1[2, :, :] = 0.5

        if nstokes > 1:
            self._atmosphere.leg_coeff.a2[2] = 3
            self._atmosphere.leg_coeff.b1[2] = -np.sqrt(6.0) / 2

        self._atmosphere.surface.albedo[:] = 0.0

        self._engine = sk.Engine(config, model_geometry, viewing_geo)

    def time_do_large(self, nlyr, calc_deriv, nstr, nstokes):  # noqa: ARG002
        self._engine.calculate_radiance(self._atmosphere)



from __future__ import annotations

from pathlib import Path

import numpy as np
import xarray as xr

import sasktran2 as sk
from sasktran2.atmosphere import Atmosphere, NativeGridDerivative
from sasktran2.optical.base import OpticalProperty, OpticalQuantities
from sasktran2.polarization import LegendreStorageView


class OpticalDatabase(OpticalProperty):
    def __init__(self, db_filepath: Path) -> None:
        """
        An optical property that is defined by a database file.  This is just a base class to handle file loading,
        derived classes must be used as actual optical properties.

        Parameters
        ----------
        db_filepath : Path
            Path to the optical database file

        Raises
        ------
        OSError
            If xarray is not installed
        """
        super().__init__()
        self._file = db_filepath

        try:
            import xarray as xr
        except ImportError:
            msg = "xarray must be installed to use OpticalDatabaseGenericAbsorber"
            raise msg from OSError

        self._database = xr.open_dataset(self._file)
        self._validate_db()


class OpticalDatabaseGenericAbsorber(OpticalDatabase):
    def __init__(self, db_filepath: Path) -> None:
        """
        A purely absorbing optical property defined by a database file.  The database must contain the following

        - xs : The absorption cross section in [m^2]

        xs must be a function of either wavelength_nm or wavenumber_cminv, and optionally temperature and pressure.

        Parameters
        ----------
        db_filepath : Path
            Path to the database file
        """
        super().__init__(db_filepath)

        self._validate_db()

    def _validate_db(self):
        # Old file format used "temperature" and "pressure" instead of the standard keys
        # "temperature_k" and "pressure_pa", if so we rename them

        if "temperature" in self._database:
            self._database = self._database.rename({"temperature": "temperature_k"})
        if "pressure" in self._database:
            self._database = self._database.rename({"pressure": "pressure_pa"})

        if "xs" not in self._database:
            msg = "xs must be defined in the optical database"
            raise ValueError(msg)

    def optical_derivatives(self, atmo: Atmosphere, **kwargs) -> dict:
        derivs = {}

        interp_handler = self._construct_interp_handler(atmo, **kwargs)

        # Split the interpolators into ones that are 'z' dependent and ones that are not

        interp_handler_z = {}
        interp_handler_noz = {}

        for key, val in interp_handler.items():
            if val[0] == "z":
                interp_handler_z[key] = val
            else:
                interp_handler_noz[key] = val

        # Get the derivatives of the cross section with respect to the z dependent variables
        partial_interp = self._database["xs"].interp(**interp_handler_noz)

        for key, val in interp_handler_z.items():
            # Interpolate over the other variables
            new_interpolants = {k: v for k, v in interp_handler_z.items() if k != key}

            partial_interp2 = partial_interp.interp(**new_interpolants)

            dT = partial_interp2.diff(key) / partial_interp2[key].diff(key)

            interp_index = np.argmax(dT[key].to_numpy() > val[1][:, np.newaxis], axis=1)

            if "z" in dT.dims:
                dT = dT.isel(
                    {
                        "z": xr.DataArray(list(range(len(interp_index))), dims="z"),
                        key: xr.DataArray(interp_index, dims="z"),
                    }
                )
            else:
                dT = dT.isel({key: interp_index})

            derivs[key] = NativeGridDerivative(d_extinction=dT.to_numpy())
            derivs[key].d_extinction[np.isnan(derivs[key].d_extinction)] = 0

        return derivs

    def _construct_interp_handler(self, atmo: Atmosphere, **kwargs) -> dict:
        coords = self._database["xs"].coords

        interp_handler = {}

        # TODO: this could probably be refactored to iterate over atmosphere properties?
        if "temperature_k" in coords:
            if atmo.temperature_k is None:
                msg = "temperature_k must be specified in Atmosphere to use OpticalDatabaseGenericAbsorber"
                raise ValueError(msg)

            interp_handler["temperature_k"] = ("z", atmo.temperature_k)

        if "pressure_pa" in coords:
            if atmo.pressure_pa is None:
                msg = "pressure_pa must be specified in Atmosphere to use OpticalDatabaseGenericAbsorber"
                raise ValueError(msg)

            interp_handler["pressure_pa"] = ("z", atmo.pressure_pa)

        if "wavelength_nm" in coords:
            if atmo.wavelengths_nm is None:
                msg = "wavelengths_nm must be specified in Atmosphere to use OpticalDatabaseGenericAbsorber"
            interp_handler["wavelength_nm"] = atmo.wavelengths_nm

        if "wavenumber_cminv" in coords:
            if atmo.wavenumbers_cminv is None:
                msg = "wavenumber_cminv must be specified in Atmosphere to use OpticalDatabaseGenericAbsorber"
            interp_handler["wavenumber_cminv"] = atmo.wavenumbers_cminv

        return interp_handler

    def atmosphere_quantities(self, atmo: sk.Atmosphere, **kwargs) -> OpticalQuantities:
        quants = OpticalQuantities(ssa=np.zeros_like(atmo.storage.ssa))

        interp_handler = self._construct_interp_handler(atmo, **kwargs)

        quants.extinction = self._database["xs"].interp(**interp_handler).to_numpy()

        # Out of bounds, set to 0
        quants.extinction[np.isnan(quants.extinction)] = 0

        return quants


class OpticalDatabaseGenericScatterer(OpticalDatabase):
    def __init__(self, db_filepath: Path) -> None:
        """
        A purely scattering optical property defined by a database file.  The database must contain the following

        - xs_total : The total cross section in [m^2]
        - xs_scattering : The scattering cross section in [m^2]
        - lm_a1 : the legendre coefficients for the phase function

        All variables must be a function of either wavelength_nm or wavenumber_cminv, and optionally any other dimension such
        as particle size.

        Parameters
        ----------
        db_filepath : Path
            Path to the database file
        """
        super().__init__(db_filepath)

        self._validate_db()

    def _validate_db(self):
        self._database["lm_a1"] /= self._database["lm_a1"].isel(legendre=0)

    def _construct_interp_handler(self, atmo: Atmosphere, **kwargs) -> dict:
        coords = self._database["xs_total"].coords
        interp_handler = {}

        if "wavelength_nm" in coords:
            if atmo.wavelengths_nm is None:
                msg = "wavelengths_nm must be specified in Atmosphere to use OpticalDatabaseGenericScatterer"
                raise ValueError(msg)
            interp_handler["wavelength_nm"] = atmo.wavelengths_nm

        if "wavenumber_cminv" in coords:
            if atmo.wavenumber_cminv is None:
                msg = "wavenumber_cminv must be specified in Atmosphere to use OpticalDatabaseGenericScatterer"
                raise ValueError(msg)
            interp_handler["wavenumber_cminv"] = atmo.wavenumber_cminv

        for name, vals in kwargs.items():
            if name in coords:
                interp_handler[name] = ("z", vals)

        return interp_handler

    def cross_sections(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> OpticalQuantities:
        quants = OpticalQuantities()

        coords = self._database["xs_total"].coords
        interp_handler = {}

        interp_handler["wavelength_nm"] = wavelengths_nm

        for name, vals in kwargs.items():
            if name in coords:
                interp_handler[name] = ("z", vals)

        ds_interp = self._database.interp(**interp_handler)

        quants.extinction = ds_interp["xs_total"].to_numpy()
        quants.ssa = ds_interp["xs_scattering"].to_numpy() / quants.extinction

        quants.extinction[np.isnan(quants.extinction)] = 0
        quants.ssa[np.isnan(quants.ssa)] = 0

        return quants

    def atmosphere_quantities(self, atmo: Atmosphere, **kwargs) -> OpticalQuantities:
        quants = OpticalQuantities()

        interp_handler = self._construct_interp_handler(atmo, **kwargs)

        num_assign_legendre = min(
            atmo.leg_coeff.a1.shape[0], len(self._database["legendre"])
        )

        if atmo.nstokes == 1:
            drop_vars = ["lm_a2", "lm_a3", "lm_a4", "lm_b1", "lm_b2"]
        elif atmo.nstokes == 3:
            drop_vars = ["lm_a4", "lm_b2"]
        else:
            drop_vars = []

        ds_interp = (
            self._database.isel(legendre=slice(0, num_assign_legendre))
            .drop(drop_vars)
            .interp(**interp_handler)
        )

        if "z" not in ds_interp.dims:
            # Our dataset has no z dependence
            ds_interp = ds_interp.expand_dims(
                dim={"z": len(atmo.model_geometry.altitudes())}
            )

        quants.extinction = np.copy(
            ds_interp["xs_total"].transpose("z", "wavelength_nm").to_numpy()
        )
        quants.ssa = np.copy(
            ds_interp["xs_scattering"].transpose("z", "wavelength_nm").to_numpy()
        )

        quants.leg_coeff = np.zeros_like(atmo.storage.leg_coeff)

        leg_coeff = LegendreStorageView(quants.leg_coeff, atmo.nstokes)

        leg_coeff.a1[:] = (
            (ds_interp["lm_a1"])
            .transpose("legendre", "z", "wavelength_nm")
            .to_numpy()[:num_assign_legendre]
        )

        # Renormalize leg_coeffs so a1 is always 1
        leg_coeff.a1[:] /= quants.leg_coeff[0, :, :][np.newaxis, :, :]

        quants.extinction[np.isnan(quants.extinction)] = 0
        quants.ssa[np.isnan(quants.ssa)] = 0

        if atmo.nstokes == 3:
            # TODO: add pol properties

            leg_coeff.a2[:] = (
                (ds_interp["lm_a2"])
                .transpose("legendre", "z", "wavelength_nm")
                .to_numpy()[:num_assign_legendre]
            )

            leg_coeff.a3[:] = (
                (ds_interp["lm_a3"])
                .transpose("legendre", "z", "wavelength_nm")
                .to_numpy()[:num_assign_legendre]
            )

            leg_coeff.b1[:] = (
                (ds_interp["lm_b1"])
                .transpose("legendre", "z", "wavelength_nm")
                .to_numpy()[:num_assign_legendre]
            )

        quants.leg_coeff[np.isnan(quants.leg_coeff)] = 0

        return quants

    def optical_derivatives(self, atmo: Atmosphere, **kwargs) -> dict:
        derivs = {}

        interp_handler = self._construct_interp_handler(atmo, **kwargs)

        # Split the interpolators into ones that are 'z' dependent and ones that are not

        interp_handler_z = {}
        interp_handler_noz = {}

        for key, val in interp_handler.items():
            if val[0] == "z":
                interp_handler_z[key] = val
            else:
                interp_handler_noz[key] = val
        num_assign_legendre = min(
            atmo.leg_coeff.a1.shape[0], len(self._database["legendre"])
        )

        if atmo.nstokes == 1:
            drop_vars = ["lm_a2", "lm_a3", "lm_a4", "lm_b1", "lm_b2"]
        elif atmo.nstokes == 3:
            drop_vars = ["lm_a4", "lm_b2"]
        else:
            drop_vars = []

        # Get the derivatives of the cross section with respect to the z dependent variables
        partial_interp = (
            self._database.isel(legendre=slice(0, num_assign_legendre))
            .drop(drop_vars)
            .interp(**interp_handler_noz)
        )

        for key, val in interp_handler_z.items():
            # If the db only contains one element in the dimension we can't take a derivative
            if len(self._database[key]) == 1:
                continue

            # Interpolate over the other variables
            new_interpolants = {k: v for k, v in interp_handler_z.items() if k != key}

            partial_interp2 = partial_interp.interp(**new_interpolants)

            dT = partial_interp2.diff(key) / partial_interp2[key].diff(key)

            interp_index = np.argmax(dT[key].to_numpy() > val[1][:, np.newaxis], axis=1)

            if "z" in dT.dims:
                dT = dT.isel(
                    {
                        "z": xr.DataArray(list(range(len(interp_index))), dims="z"),
                        key: xr.DataArray(interp_index, dims="z"),
                    }
                )
            else:
                dT = dT.isel({key: xr.DataArray(interp_index, dims="z")})

            derivs[key] = NativeGridDerivative(
                d_extinction=dT["xs_total"].transpose("z", "wavelength_nm").to_numpy(),
                d_ssa=dT["xs_scattering"].transpose("z", "wavelength_nm").to_numpy(),
                d_leg_coeff=np.zeros_like(atmo.storage.leg_coeff),
            )

            d_leg_coeff = LegendreStorageView(derivs[key].d_leg_coeff, atmo.nstokes)

            d_leg_coeff.a1[:] = (
                dT["lm_a1"]
                .transpose("legendre", "z", "wavelength_nm")
                .to_numpy()[:num_assign_legendre]
            )

            if atmo.nstokes == 3:
                d_leg_coeff.a2[:] = (
                    dT["lm_a2"]
                    .transpose("legendre", "z", "wavelength_nm")
                    .to_numpy()[:num_assign_legendre]
                )
                d_leg_coeff.a3[:] = (
                    dT["lm_a3"]
                    .transpose("legendre", "z", "wavelength_nm")
                    .to_numpy()[:num_assign_legendre]
                )
                d_leg_coeff.b1[:] = (
                    dT["lm_b1"]
                    .transpose("legendre", "z", "wavelength_nm")
                    .to_numpy()[:num_assign_legendre]
                )

            derivs[key].d_extinction[np.isnan(derivs[key].d_extinction)] = 0
            derivs[key].d_ssa[np.isnan(derivs[key].d_ssa)] = 0
            derivs[key].d_leg_coeff[np.isnan(derivs[key].d_leg_coeff)] = 0

        return derivs

    def cross_section_derivatives(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> dict:
        derivs = {}
        coords = self._database["xs_total"].coords
        interp_handler = {}

        interp_handler["wavelength_nm"] = wavelengths_nm

        for name, vals in kwargs.items():
            if name in coords:
                interp_handler[name] = ("z", vals)

        # Split the interpolators into ones that are 'z' dependent and ones that are not

        interp_handler_z = {}
        interp_handler_noz = {}

        for key, val in interp_handler.items():
            if val[0] == "z":
                interp_handler_z[key] = val
            else:
                interp_handler_noz[key] = val

        # Get the derivatives of the cross section with respect to the z dependent variables
        partial_interp = self._database["xs_total"].interp(**interp_handler_noz)

        for key, val in interp_handler_z.items():
            # If the db only contains one element in the dimension we can't take a derivative
            if len(self._database[key]) == 1:
                continue

            # Interpolate over the other variables
            new_interpolants = {k: v for k, v in interp_handler_z.items() if k != key}

            partial_interp2 = partial_interp.interp(**new_interpolants)

            dT = partial_interp2.diff(key) / partial_interp2[key].diff(key)

            interp_index = np.argmax(dT[key].to_numpy() > val[1][:, np.newaxis], axis=1)

            if "z" in dT.dims:
                dT = dT.isel(
                    {
                        "z": xr.DataArray(list(range(len(interp_index))), dims="z"),
                        key: xr.DataArray(interp_index, dims="z"),
                    }
                )
            else:
                dT = dT.isel({key: xr.DataArray(interp_index, dims="z")})

            derivs[key] = dT.to_numpy().flatten()

        return derivs



import sasktran as sk_old
import sasktran2 as sk
import numpy as np
from tqdm import tqdm
import xarray as xr
from pathlib import Path


OUT_DIR = Path(
    "/Users/dannyz/OneDrive - University of Saskatchewan/sasktran_databases/dist/cross_sections/"
)


PRESSURE_GRID = (
    np.array(
        [
            1.03181655e03,
            9.99910226e02,
            9.72231997e02,
            9.44960640e02,
            9.18098202e02,
            8.91646636e02,
            8.65607893e02,
            8.39983721e02,
            8.14775620e02,
            7.89985094e02,
            7.65613488e02,
            7.41661804e02,
            7.18131094e02,
            6.95022257e02,
            6.72335891e02,
            6.50072549e02,
            6.28232578e02,
            6.06816233e02,
            5.85823558e02,
            5.65254408e02,
            5.45108580e02,
            5.25385676e02,
            5.06085044e02,
            4.87205987e02,
            4.68747652e02,
            4.50708890e02,
            4.33088502e02,
            4.15885138e02,
            3.99097246e02,
            3.82723078e02,
            3.66760783e02,
            3.51208311e02,
            3.36063513e02,
            3.21323986e02,
            3.06987184e02,
            2.93050454e02,
            2.79510897e02,
            2.66365512e02,
            2.53611100e02,
            2.41244361e02,
            2.29261844e02,
            2.17659797e02,
            2.06434425e02,
            1.95581772e02,
            1.85097641e02,
            1.74977731e02,
            1.65217644e02,
            1.55812775e02,
            1.46758379e02,
            1.38049500e02,
            1.29681042e02,
            1.21647854e02,
            1.13944585e02,
            1.06565734e02,
            9.95056250e01,
            9.27584798e01,
            8.63183864e01,
            8.01792901e01,
            7.43350047e01,
            6.87792048e01,
            6.35054499e01,
            5.85071744e01,
            5.37776926e01,
            4.93102089e01,
            4.50978073e01,
            4.11334673e01,
            3.74100678e01,
            3.39203935e01,
            3.06571276e01,
            2.76128649e01,
            2.47801240e01,
            2.21513493e01,
            1.97189141e01,
            1.74751173e01,
            1.54122078e01,
            1.35223946e01,
            1.17978408e01,
            1.02306666e01,
            8.81297618e00,
            7.53686238e00,
            6.39440891e00,
            5.37770790e00,
            4.47887025e00,
            3.69003722e00,
            3.00339331e00,
            2.41118145e00,
            1.90571877e00,
            1.47941371e00,
            1.12478403e00,
            8.34477191e-01,
            6.01292929e-01,
            4.18208541e-01,
            2.78406802e-01,
            1.75308171e-01,
            1.02608084e-01,
            5.43204621e-02,
        ]
    )
    * 100
)
TEMP_GRID = np.arange(190, 311, 10)
HIRES_WAVEL_GRID = np.arange(200.0, 1200.0, 0.001)

NUM_COMBINE = 100

MAX_WAVEL_RES = 0.1
SIGNIFICANT_VOD_THRESHOLD = 0.01


def typical_vod(species):
    const = sk.climatology.mipas.constituent(
        species, None, dataset="fascode", climatology="std"
    )

    # Hacks for P/T
    temperature_k = (
        sk.climatology.mipas.constituent(
            "TEM", None, dataset="fascode", climatology="std"
        ).vmr
        * 1e6
    )
    pressure_pa = (
        sk.climatology.mipas.constituent(
            "PRE", None, dataset="fascode", climatology="std"
        ).vmr
        * 1e6
        * 1e2
    )
    alts_m = (
        sk.climatology.mipas.constituent("HGT", None, "fascode", "std").vmr * 1e6 * 1e3
    )

    nden = sk.optical.pressure_temperature_to_numberdensity(pressure_pa, temperature_k)

    integrated_nden = np.trapz(nden, alts_m)

    return integrated_nden


def reduced_wv():
    all_wv = []

    i = 0
    while i + NUM_COMBINE < len(HIRES_WAVEL_GRID):
        all_wv.append(np.nanmean(HIRES_WAVEL_GRID[i : i + NUM_COMBINE]))

        i += NUM_COMBINE

    all_wv = np.array(all_wv)

    return all_wv


def reduce_xs(vod, xs):
    all_wv = []
    all_xs = []

    i = 0
    while i + NUM_COMBINE < len(HIRES_WAVEL_GRID):
        all_xs.append(np.nanmean(xs[i : i + NUM_COMBINE]))

        i += NUM_COMBINE

    all_xs = np.array(all_xs)

    return all_xs


def gen_db(species, ltol=1e-9):
    vod = typical_vod(species)

    r_wv = reduced_wv()
    reduced_xs = np.zeros((len(PRESSURE_GRID), len(TEMP_GRID), len(r_wv)))

    for idx, pres in tqdm(enumerate(PRESSURE_GRID)):
        for idy, temp in enumerate(TEMP_GRID):
            xs = (
                sk_old.HITRANChemical(species, line_tolerance=ltol)
                .calculate_cross_sections(
                    sk_old.ClimatologyUserDefined(
                        altitudes=[0, 100000],
                        values={
                            "SKCLIMATOLOGY_PRESSURE_PA": [pres, pres],
                            "SKCLIMATOLOGY_TEMPERATURE_K": [temp, temp],
                        },
                    ),
                    0,
                    0,
                    10000,
                    54372,
                    HIRES_WAVEL_GRID,
                )
                .total
                / 1e4
            )

            reduced_xs[idx, idy] = reduce_xs(vod, xs)

            # xs = sk_old.HITRANChemical(species, line_tolerance=0).calculate_cross_sections(sk_old.ClimatologyUserDefined(altitudes=[0, 100000], values={"SKCLIMATOLOGY_PRESSURE_PA": [pres, pres], "SKCLIMATOLOGY_TEMPERATURE_K": [temp, temp]}), 0, 0, 10000, 54372, HIRES_WAVEL_GRID).total / 1e4
            #
            # reduced_wv2, reduced_xs2 = reduce_xs(vod, xs)

    ds = xr.Dataset(
        {"xs": (["pressure", "temperature", "wavelength_nm"], reduced_xs)},
        coords={
            "pressure": PRESSURE_GRID,
            "temperature": TEMP_GRID,
            "wavelength_nm": r_wv,
        },
    )

    out_dir = OUT_DIR.joinpath(species.lower())
    if not out_dir.exists():
        out_dir.mkdir()

    ds.to_netcdf(out_dir.joinpath("hitran_01nm_res.nc"))


gen_db("O2")
# gen_db("H2O")



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
import sasktran2 as sk


def test_limb_from_tangent_parameters():
    pytest.importorskip("astropy")
    solar = sk.solar.SolarGeometryHandlerAstropy()

    geo = sk.viewinggeo.LimbVertical.from_tangent_parameters(
        solar,
        np.arange(0, 65000, 1000.0),
        20,
        30,
        pd.Timestamp("2024-11-12 20:00:00"),
        200000,
        0,
    )

    _ = geo.model_geometry(np.arange(0, 65000, 1000.0))


def test_chain_with_viewing_container():
    pytest.importorskip("astropy")
    solar = sk.solar.SolarGeometryHandlerAstropy()

    geo = sk.viewinggeo.LimbVertical.from_tangent_parameters(
        solar,
        np.arange(0, 65000, 1000.0),
        20,
        30,
        pd.Timestamp("2024-11-12 20:00:00"),
        200000,
        0,
    )

    model_geo = geo.model_geometry(np.arange(0, 65000, 1000.0))

    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geo, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    atmosphere["ozone"] = sk.constituent.VMRAltitudeAbsorber(
        sk.optical.O3DBM(),
        model_geo.altitudes(),
        np.ones_like(model_geo.altitudes()) * 1e-6,
    )

    engine = sk.Engine(config, model_geo, geo)

    radiance = engine.calculate_radiance(atmosphere)

    assert "tangent_altitude" in radiance



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def test_fascode():
    fascode_valid = [
        "H2O",
        "CO2",
        "O3",
        "N2O",
        "CO",
        "CH4",
        "O2",
        "NO",
        "SO2",
        "NO2",
        "NH3",
        "HNO3",
        "OH",
        "HF",
        "HCl",
        "HBr",
        "HI",
        "ClO",
        "OCS",
        "H2CO",
        "HOCl",
        "N2",
        "HCN",
        "CH3Cl",
        "H2O2",
        "C2H2",
        "C2H6",
        "PH3",
        "COF2",
        "SF6",
        "H2S",
        "CFCl3",
        "CF2Cl2",
        "CClF3",
        "CF4",
        "CHCl2F",
        "CHClF2",
        "C2Cl3F3",
        "C2Cl2F4",
        "C2ClF5",
        "CCl4",
        "ClONO2",
        "N2O5",
        "HNO4",
    ]

    fascode_states = ["tro", "mls", "mlw", "sas", "saw", "std"]

    for fascode in fascode_valid:
        for fascode_state in fascode_states:
            sk.climatology.mipas.constituent(
                fascode, None, dataset="fascode", climatology=fascode_state
            )


def test_mipas1998():
    mipas_valid = [
        "N2",
        "O2",
        "O3P",
        "CO2",
        "O3",
        "H2O",
        "CH4",
        "N2O",
        "HNO3",
        "CO",
        "NO2",
        "N2O5",
        "ClO",
        "HOCl",
        "ClONO2",
        "NO",
        "HCN",
        "H2O2",
        "F12",
        "F14",
        "F22",
        "COF2",
        "OCS",
        "NH3",
        "SO2",
        "CFCl3",
        "C2H2",
        "C2H6",
        "CCl4",
        "SF6",
        "HNO4",
        "CH3Cl",
        "CClF3",
        "CHCl2F",
        "C2Cl3F3",
        "C2Cl2F4",
    ]

    mipas_states = ["day_imk", "ngt_imk", "win_imk", "sum_imk"]

    for mipas in mipas_valid:
        for mipas_state in mipas_states:
            sk.climatology.mipas.constituent(
                mipas, None, dataset="mipas_1998", climatology=mipas_state
            )


def test_mipas2001():
    mipas_valid = [
        "N2",
        "O2",
        "CO2",
        "O3",
        "H2O",
        "CH4",
        "N2O",
        "HNO3",
        "CO",
        "NO2",
        "N2O5",
        "ClO",
        "HOCl",
        "ClONO2",
        "NO",
        "HNO4",
        "HCN",
        "NH3",
        "F11",
        "F12",
        "F14",
        "F22",
        "CCl4",
        "COF2",
        "H2O2",
        "C2H2",
        "C2H6",
        "OCS",
        "SO2",
        "SF6",
        "CClF3",
        "CHCl2F",
        "C2Cl3F3",
        "C2Cl2F4",
        "C2ClF5",
        "CH3Cl",
        "H2S",
    ]

    mipas_states = ["day", "ngt", "win", "sum", "equ"]

    for mipas in mipas_valid:
        for mipas_state in mipas_states:
            sk.climatology.mipas.constituent(
                mipas, None, dataset="mipas_2001", climatology=mipas_state
            )


def test_add_to_atmosphere():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for alt in [10000, 20000, 30000, 40000]:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    wavel = np.arange(280.0, 800.0, 10)
    atmosphere = sk.Atmosphere(model_geometry, config, wavelengths_nm=wavel)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    sk.climatology.mipas.add_to_atmosphere(
        atmosphere, {"O3": sk.optical.O3DBM(), "NO2": sk.optical.NO2Vandaele()}
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)

    _ = engine.calculate_radiance(atmosphere)



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def _test_scenarios():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 4
    config.delta_m_scaling = False

    altitude_grid = np.arange(0, 65001, 1000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geo.add_ray(sk.TangentAltitudeSolar(tan_alt, 0, 600000, 0.6))

    wavel = np.array([310, 330, 350, 600])

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    scen = []

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    return scen


def test_lambertian_surface_constant():
    """
    Checks that we can create a lambertian surface that is constant in wavelength
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.LambertianSurface(0.3)

        _ = engine.calculate_radiance(scen["atmosphere"])

        np.testing.assert_allclose(scen["atmosphere"].surface.albedo, 0.3)


def test_lambertian_surface_interp_wavelength():
    """
    Checks that we can create a lambertian surface that is interpolated in wavelength
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.LambertianSurface(
            albedo=[0.1, 0.3], wavelengths_nm=[280, 600]
        )

        _ = engine.calculate_radiance(scen["atmosphere"])

        np.testing.assert_allclose(
            scen["atmosphere"].surface.albedo.flatten(),
            [0.11875, 0.13125, 0.14375, 0.3],
        )


def test_lambertian_surface_interp_wavenumber():
    """
    Checks that we can create a lambertian surface that is interpolated in wavenumber
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.LambertianSurface(
            albedo=[0.1, 0.3], wavenumbers_cminv=[10000, 30000]
        )

        _ = engine.calculate_radiance(scen["atmosphere"])

        np.testing.assert_allclose(
            scen["atmosphere"].surface.albedo.flatten(), [0, 0, 0.28571429, 0.16666667]
        )


def test_lambertian_surface_native():
    """
    Checks that we can create a lambertian surface that is on the same grid as the atmospheric wavelength
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.LambertianSurface(
            albedo=[0.1, 0.8, 0.9, 0.3]
        )

        _ = engine.calculate_radiance(scen["atmosphere"])

        np.testing.assert_allclose(
            scen["atmosphere"].surface.albedo.flatten(), [0.1, 0.8, 0.9, 0.3]
        )


def test_lambertian_surface_wf_native():
    """
    Checks that the lambertian surface weighting functions are correct when specified on the native atmospheric wavelength grid
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.LambertianSurface(
            albedo=[0.1, 0.8, 0.9, 0.3]
        )

        radiance = sk.test_util.wf.numeric_wf(
            scen["atmosphere"]["surface"].albedo,
            0.001,
            engine,
            scen["atmosphere"],
            "wf_surface_albedo",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_surface_albedo"],
            radiance["wf_surface_albedo_numeric"],
            "surface_wavelength",
            decimal=6,
        )


def test_lambertian_surface_wf_scalar():
    """
    Checks that the lambertian surface weighting functions are correct for a scalar albedo
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.LambertianSurface(albedo=0.3)

        radiance = sk.test_util.wf.numeric_wf(
            scen["atmosphere"]["surface"].albedo,
            0.001,
            engine,
            scen["atmosphere"],
            "wf_surface_albedo",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_surface_albedo"],
            radiance["wf_surface_albedo_numeric"],
            "surface_wavelength",
            decimal=6,
        )


def test_lambertian_surface_wf_interpolated():
    """
    Checks that the lambertian surface weighting functions are correct when we are interpolating over wavelength
    """
    scens = _test_scenarios()

    for scen in scens:
        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        scen["atmosphere"]["surface"] = sk.constituent.LambertianSurface(
            albedo=[0.1, 0.3], wavelengths_nm=[280, 600]
        )

        radiance = sk.test_util.wf.numeric_wf(
            scen["atmosphere"]["surface"].albedo,
            0.001,
            engine,
            scen["atmosphere"],
            "wf_surface_albedo",
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_surface_albedo"],
            radiance["wf_surface_albedo_numeric"],
            "surface_wavelength",
            decimal=6,
        )



from __future__ import annotations

from pathlib import Path

import xarray as xr

from sasktran2.util import get_hapi

from .base import CachedDatabase


class HITRANLineDatabase(CachedDatabase):
    def __init__(
        self, db_root: Path | None = None, rel_path: Path | None = "hitran_lines"
    ):
        """
        Downloads the HITRAN Line databases from hitran using the hitran-api package.  Both the raw
        HITRAN data is stored for use in the hitran-api, and a netcdf file version for use in the
        SASKTRAN2 broadening routines.
        """
        CachedDatabase.__init__(self, db_root, rel_path=rel_path)

    def path(self, key: str, **kwargs) -> Path:
        data_file = self._db_root.joinpath(f"{key}.data")
        header_file = self._db_root.joinpath(f"{key}.header")
        if not (data_file.exists() and header_file.exists()):
            # download lines for this molecule
            self._download_line_db(key)

    def load_ds(self, key: str, **kwargs) -> xr.Dataset:
        nc_file = self._db_root.joinpath(f"{key}.nc")
        if nc_file.exists():
            return xr.open_dataset(nc_file)

        hapi = get_hapi()

        self.path(key)
        self.initialize_hapi(key)

        data_vars = {}
        for param in hapi.PARLIST_ALL:
            try:
                param_column = hapi.getColumn(key, param)
                data_vars[param] = ("line", param_column)
            except KeyError:  # param does not exist for this molecule
                continue

        ds = xr.Dataset(data_vars=data_vars)
        ds.to_netcdf(nc_file)
        return ds

    def clear(self):
        # deletes all .data and .header files in database folder
        for header_file in self._db_root.glob("*.header"):
            header_file.unlink()
        for data_file in self._db_root.glob("*.data"):
            data_file.unlink()

    def _download_line_db(self, molecule):
        hapi = get_hapi()

        self.initialize_hapi()

        # Get global isotopologue ids for this molecule
        mol_index = hapi.ISO_ID_INDEX["mol_name"]
        iso_ids = []
        for id in hapi.ISO_ID:
            if hapi.ISO_ID[id][mol_index] == molecule:
                iso_ids.append(id)

        # Download files
        hapi.fetch_by_ids(
            molecule, iso_ids, 0.0, 1.0e6, ParameterGroups=["par_line", "sdvoigt", "ht"]
        )

        return

    def initialize_hapi(self, table_name: str | None = None):
        """
        Initialize the HAPI database with one or no tables. This is to reduce the time and memory that hapi.db_begin
        uses when it is given a folder with many data files.
        """
        hapi = get_hapi()

        hapi.VARIABLES["BACKEND_DATABASE_NAME"] = str(self._db_root)
        hapi.LOCAL_TABLE_CACHE = {}
        if table_name is not None:
            hapi.storage2cache(table_name)

        return



from __future__ import annotations

import numpy as np


class LegendreStorageView:
    def __init__(self, leg_coeff_stacked: np.ndarray, nstokes: int):
        """
        Convenience class to allow easier interfacing with the stacked Legendre coefficient format used
        internally in SASKTRAN2 for performance reasons.

        Internally SASKTRAN2 contains an array of Legendre coefficient (ldim, wavelength, geometric location),
        where ldim is a stacked Legendre dimension.  In the scalar case this is simply the Legendre coefficients
        :math:`\beta_l`, but in the `nstokes=3` case this dimension is :math:`a_{1,1}, a_{2, 1}, a_{3, 1}, b_{1, 1}, a_{1, 2},...`
        where the second coefficient is the expansion index.

        This class provides accessors to set/view :math:`a_1, a_2, a_3, b_1`.


        Parameters
        ----------
        leg_coeff_stacked : np.ndarray
            The raw stacked legendre coefficient data
        nstokes : int
            The number of stokes parameters being included.  Currently only 1 and 3 are supported.

        Raises
        ------
        ValueError
            If nstokes is not 1 or 3
        """
        if nstokes == 1:
            num_coeff = 1
        elif nstokes == 3:
            num_coeff = 4
        else:
            msg = "Only nstokes=1 and nstokes=3 are supported by LegendreStorageView"
            raise ValueError(msg)

        self._raw = leg_coeff_stacked.view()

        self._a1 = self._raw[::num_coeff].view()

        if nstokes == 3:
            self._a2 = self._raw[1::num_coeff].view()
            self._a3 = self._raw[2::num_coeff].view()
            self._b1 = self._raw[3::num_coeff].view()
        else:
            self._a2 = None
            self._a3 = None
            self._b1 = None

    @property
    def a1(self):
        return self._a1

    @property
    def a2(self):
        return self._a2

    @property
    def a3(self):
        return self._a3

    @property
    def b1(self):
        return self._b1



from __future__ import annotations

import numpy as np
from sasktran2.optical.refraction import ciddor_index_of_refraction


def test_ciddor():
    # Check with 0 specific humidity
    n1 = ciddor_index_of_refraction(
        temperatures_k=273.15 + 20,
        pressure_pa=101325,
        specific_humidity=0,
        x_co2=450,
        wavelength_nm=633,
    )
    np.testing.assert_almost_equal(n1, 1.0002718)

    # Adjust pressure
    n2 = ciddor_index_of_refraction(
        temperatures_k=273.15 + 20,
        pressure_pa=60000,
        specific_humidity=0,
        x_co2=450,
        wavelength_nm=633,
    )
    np.testing.assert_almost_equal(n2, 1.000160924)

    # Adjust temperature
    n3 = ciddor_index_of_refraction(
        temperatures_k=273.15 + 50,
        pressure_pa=100000,
        specific_humidity=0,
        x_co2=450,
        wavelength_nm=633,
    )
    np.testing.assert_almost_equal(n3, 1.000243285)

    # Adjust wavelength
    n4 = ciddor_index_of_refraction(
        temperatures_k=273.15 + 20,
        pressure_pa=101325,
        specific_humidity=0,
        x_co2=450,
        wavelength_nm=1700,
    )
    np.testing.assert_almost_equal(n4, 1.000268479)

    # Adjust specific humidity
    # RH of 100% gives saturation vapor pressure of 2339 Pa at 20 C
    # Enhancement factor
    p = 100e3
    f = 1.00062 + 3.14e-8 * p + 5.60e-7 * (20**2)
    xv = f * 2339 / p

    # Convert to specific humidity
    q = xv / (xv + (1 - xv) * (18.01528 / 28.9647))

    n5 = ciddor_index_of_refraction(
        temperatures_k=273.15 + 20,
        pressure_pa=p,
        specific_humidity=q,
        x_co2=450,
        wavelength_nm=633,
    )
    np.testing.assert_allclose(n5, 1.000267394)



from __future__ import annotations

import hashlib
import json
import logging
from pathlib import Path

import numpy as np
import xarray as xr

from sasktran2.optical.database import OpticalDatabaseGenericAbsorber
from sasktran2.optical.hitran import HITRANAbsorber
from sasktran2.units import wavenumber_cminv_to_wavlength_nm
from sasktran2.util import get_hapi

from .base import CachedDatabase
from .hitran_line import HITRANLineDatabase

PRESSURE_GRID = (
    np.array(
        [
            1.03181655e03,
            9.99910226e02,
            9.72231997e02,
            9.44960640e02,
            9.18098202e02,
            8.91646636e02,
            8.65607893e02,
            8.39983721e02,
            8.14775620e02,
            7.89985094e02,
            7.65613488e02,
            7.41661804e02,
            7.18131094e02,
            6.95022257e02,
            6.72335891e02,
            6.50072549e02,
            6.28232578e02,
            6.06816233e02,
            5.85823558e02,
            5.65254408e02,
            5.45108580e02,
            5.25385676e02,
            5.06085044e02,
            4.87205987e02,
            4.68747652e02,
            4.50708890e02,
            4.33088502e02,
            4.15885138e02,
            3.99097246e02,
            3.82723078e02,
            3.66760783e02,
            3.51208311e02,
            3.36063513e02,
            3.21323986e02,
            3.06987184e02,
            2.93050454e02,
            2.79510897e02,
            2.66365512e02,
            2.53611100e02,
            2.41244361e02,
            2.29261844e02,
            2.17659797e02,
            2.06434425e02,
            1.95581772e02,
            1.85097641e02,
            1.74977731e02,
            1.65217644e02,
            1.55812775e02,
            1.46758379e02,
            1.38049500e02,
            1.29681042e02,
            1.21647854e02,
            1.13944585e02,
            1.06565734e02,
            9.95056250e01,
            9.27584798e01,
            8.63183864e01,
            8.01792901e01,
            7.43350047e01,
            6.87792048e01,
            6.35054499e01,
            5.85071744e01,
            5.37776926e01,
            4.93102089e01,
            4.50978073e01,
            4.11334673e01,
            3.74100678e01,
            3.39203935e01,
            3.06571276e01,
            2.76128649e01,
            2.47801240e01,
            2.21513493e01,
            1.97189141e01,
            1.74751173e01,
            1.54122078e01,
            1.35223946e01,
            1.17978408e01,
            1.02306666e01,
            8.81297618e00,
            7.53686238e00,
            6.39440891e00,
            5.37770790e00,
            4.47887025e00,
            3.69003722e00,
            3.00339331e00,
            2.41118145e00,
            1.90571877e00,
            1.47941371e00,
            1.12478403e00,
            8.34477191e-01,
            6.01292929e-01,
            4.18208541e-01,
            2.78406802e-01,
            1.75308171e-01,
            1.02608084e-01,
            5.43204621e-02,
        ]
    )
    * 100
)
TEMP_GRID = np.arange(190, 311, 10)


class HITRANDatabase(CachedDatabase, OpticalDatabaseGenericAbsorber):
    def __init__(
        self,
        molecule: str,
        start_wavenumber: float,
        end_wavenumber: float,
        wavenumber_resolution: float,
        reduction_factor: int,
        num_threads: int = 1,
        db_root: Path | None = None,
        backend: str = "sasktran2",
        profile: str = "voigt",
    ) -> None:
        """


        Parameters
        ----------
        molecule : str
            The HITRAN molecule identifier, e.g. "CO2", "H2O", "O3", "CH4"
        start_wavenumber : float
            The start wavenumber of the database in cm^-1
        end_wavenumber : float
            The end wavenumber of the database in cm^-1
        wavenumber_resolution : float
            The wavenumber resolution of the database in cm^-1
        num_threads: int, optional
            The number of threads to use in the cross section calculation. Default is 1.
        db_root : Path, optional
            The root directory to store the database, by default None
        backend : str, optional
            The backend to use, by default "sasktran2" which uses the internal broadening algorithms.
            options are ["sasktran2", "sasktran_legacy", "hapi"]. Default is "sasktran2".
            Note: "sasktran_legacy" requires the sasktran package to be installed.
            Both the "sasktran2" and "hapi" backends use the hapi package to access the HITRAN database.
        profile : str, optional
            The line shape profile to use in the cross section calculation. Currently, only the "hapi" backend supports options other
            than the default "voigt". Supported options are ["voigt", "sdvoigt", "ht", "priority", "lorentz", "doppler"].
            Default is "voigt".
        """

        class NumpyEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, np.ndarray):
                    return obj.tolist()
                return json.JSONEncoder.default(self, obj)

        self._profile = profile.lower()
        if backend == "sasktran_legacy" and self._profile != "voigt":
            logging.warning(
                "%s line profile is not supported by sasktran_legacy backend. Switching to voigt",
                self._profile,
            )
            self._profile = "voigt"

        hasher = hashlib.sha1()
        encoded = json.dumps(
            {
                "start_wavenumber": start_wavenumber,
                "end_wavenumber": end_wavenumber,
                "wavenumber_resolution": wavenumber_resolution,
                "reduction_factor": reduction_factor,
                "profile": self._profile,
            },
            sort_keys=True,
            cls=NumpyEncoder,
        ).encode()
        hasher.update(encoded)
        identifier = hasher.hexdigest()
        self._backend = backend
        self._start_wavenumber = start_wavenumber
        self._end_wavenumber = end_wavenumber
        self._wavenumber_resolution = wavenumber_resolution
        self._reduction_factor = reduction_factor
        self._molecule = molecule
        self._ltol = 1e-9
        self._wavenumber_wing = 50
        self._wavenumber_margin = 10
        self._num_threads = num_threads

        CachedDatabase.__init__(
            self,
            db_root=db_root,
            rel_path=Path("hitran").joinpath(self._molecule, self._backend),
        )

        self._data_file = self._db_root.joinpath(f"{identifier}.nc")

        OpticalDatabaseGenericAbsorber.__init__(self, self.path())

    def generate(self):
        if self._backend == "sasktran2":
            self._generate_sasktran2()
        elif self._backend == "sasktran_legacy":
            self._generate_sasktran_legacy()
        elif self._backend == "hapi":
            self._generate_hapi()
        else:
            msg = f"Invalid backend {self._backend}"
            raise ValueError(msg)

    def clear(self):
        if self._data_file.exists():
            self._data_file.unlink()

    def path(self, key: str | None = None, **kwargs) -> Path | None:  # noqa: ARG002
        if not self._data_file.exists():
            self.generate()
        return self._data_file

    def load_ds(self, key: str | None = None, **kwargs) -> xr.Dataset:
        return xr.open_dataset(self.path(key, **kwargs))

    def _generate_sasktran_legacy(self):
        try:
            from sasktran import ClimatologyUserDefined, HITRANChemical
        except ImportError as err:
            msg = "sasktran_legacy is required to generate Mie databases, try pip install sasktran"
            raise ImportError(msg) from err

        hires_wavenumber_grid = np.arange(
            self._start_wavenumber, self._end_wavenumber, self._wavenumber_resolution
        )

        xs = np.zeros((len(PRESSURE_GRID), len(TEMP_GRID), len(hires_wavenumber_grid)))

        for idx, pres in enumerate(PRESSURE_GRID):
            for idy, temp in enumerate(TEMP_GRID):
                xs[idx, idy] = (
                    HITRANChemical(self._molecule, line_tolerance=self._ltol)
                    .calculate_cross_sections(
                        ClimatologyUserDefined(
                            altitudes=[0, 100000],
                            values={
                                "SKCLIMATOLOGY_PRESSURE_PA": [pres, pres],
                                "SKCLIMATOLOGY_TEMPERATURE_K": [temp, temp],
                            },
                        ),
                        0,
                        0,
                        10000,
                        54372,
                        wavenumber_cminv_to_wavlength_nm(hires_wavenumber_grid),
                    )
                    .total
                    / 1e4
                )

        ds = xr.Dataset(
            {"xs": (["pressure", "temperature", "wavenumber_cminv"], xs)},
            coords={
                "pressure": PRESSURE_GRID,
                "temperature": TEMP_GRID,
                "wavenumber_cminv": hires_wavenumber_grid,
            },
        )

        ds.to_netcdf(self._data_file)

    def _generate_hapi(self):
        line_db = HITRANLineDatabase()
        line_db.path(self._molecule)

        hapi = get_hapi()

        abs_cal_fn = {
            "voigt": hapi.absorptionCoefficient_Voigt,
            "sdvoigt": hapi.absorptionCoefficient_SDVoigt,
            "ht": hapi.absorptionCoefficient_HT,
            "priority": hapi.absorptionCoefficient_Priority,
            "lorentz": hapi.absorptionCoefficient_Lorentz,
            "doppler": hapi.absorptionCoefficient_Doppler,
        }
        if self._profile not in abs_cal_fn:
            msg = f"Invalid line profile {self._profile}"
            raise ValueError(msg)

        line_db.initialize_hapi(self._molecule)

        hapi.select(
            self._molecule,
            DestinationTableName="spectral_window",
            Conditions=(
                "between",
                "nu",
                self._start_wavenumber - self._wavenumber_margin,
                self._end_wavenumber + self._wavenumber_margin,
            ),
        )

        hires_wavenumber_grid = np.arange(
            self._start_wavenumber, self._end_wavenumber, self._wavenumber_resolution
        )

        xs = np.zeros((len(PRESSURE_GRID), len(TEMP_GRID), len(hires_wavenumber_grid)))

        for idx, pres in enumerate(PRESSURE_GRID):
            for idy, temp in enumerate(TEMP_GRID):
                _, xs_hapi = abs_cal_fn[self._profile](
                    SourceTables="spectral_window",
                    Environment={"T": temp, "p": pres / 101325.0},
                    WavenumberGrid=hires_wavenumber_grid.tolist(),
                    WavenumberWing=self._wavenumber_wing,
                )
                xs[idx, idy] = xs_hapi / 1e4

        ds = xr.Dataset(
            {"xs": (["pressure", "temperature", "wavenumber_cminv"], xs)},
            coords={
                "pressure": PRESSURE_GRID,
                "temperature": TEMP_GRID,
                "wavenumber_cminv": hires_wavenumber_grid,
            },
        )

        ds.to_netcdf(self._data_file)

    def _generate_sasktran2(self):
        hires_wavenumber_grid = np.arange(
            self._start_wavenumber, self._end_wavenumber, self._wavenumber_resolution
        )

        calculator = HITRANAbsorber(
            self._molecule, line_contribution_width=self._wavenumber_wing / 2
        )
        xs = np.zeros((len(PRESSURE_GRID), len(TEMP_GRID), len(hires_wavenumber_grid)))

        for i, temperature in enumerate(TEMP_GRID):
            xs[:, i, :] = calculator.cross_sections(
                temperature_k=np.ones(len(PRESSURE_GRID)) * temperature,
                pressure_pa=PRESSURE_GRID,
                wavelengths_nm=wavenumber_cminv_to_wavlength_nm(hires_wavenumber_grid),
                num_threads=self._num_threads,
                altitudes_m=np.ones(len(PRESSURE_GRID)) * 0,
            ).T
        ds = xr.Dataset(
            {"xs": (["pressure", "temperature", "wavenumber_cminv"], xs)},
            coords={
                "pressure": PRESSURE_GRID,
                "temperature": TEMP_GRID,
                "wavenumber_cminv": hires_wavenumber_grid,
            },
        )

        ds.to_netcdf(self._data_file)



from __future__ import annotations

import numpy as np
import sasktran2 as sk


def _test_scenarios():
    config = sk.Config()
    config.multiple_scatter_source = sk.MultipleScatterSource.DiscreteOrdinates
    config.num_streams = 2

    altitude_grid = np.arange(0, 65001, 1000.0)

    geometry = sk.Geometry1D(
        0.6,
        0,
        6327000,
        altitude_grid,
        sk.InterpolationMethod.LinearInterpolation,
        sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    for tan_alt in np.arange(10000, 60000, 2000):
        viewing_geo.add_ray(sk.TangentAltitudeSolar(tan_alt, 0, 600000, 0.6))

    wavel = np.array([310, 330, 350, 600])

    atmosphere = sk.Atmosphere(geometry, config, wavelengths_nm=wavel)

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()

    scen = []

    scen.append(
        {
            "config": config,
            "geometry": geometry,
            "viewing_geo": viewing_geo,
            "atmosphere": atmosphere,
        }
    )

    return scen


def test_o2o2_construction():
    """
    Test that the CollisionInducedAbsorber class can be constructed.
    """
    sk.constituent.CollisionInducedAbsorber(
        sk.optical.HITRANCollision("O2O2"), name="O2O2"
    )


def test_o2o2_wf_temperature():
    """
    Tests that the CollisionInducedAbsorber class calculates derivatives with respect to temperature correctly.
    """

    scens = _test_scenarios()

    for scen in scens:
        atmosphere = scen["atmosphere"]

        atmosphere["o2o2"] = sk.constituent.CollisionInducedAbsorber(
            sk.optical.HITRANCollision("O2O2"), name="O2O2"
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere.temperature_k, 0.00001, engine, atmosphere, "wf_temperature_k"
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_temperature_k"],
            radiance["wf_temperature_k_numeric"],
            wf_dim="altitude",
            decimal=4,
        )


def test_o2o2_wf_pressure():
    """
    Tests that the CollisionInducedAbsorber class calculates derivatives with respect to pressure correctly.
    """

    scens = _test_scenarios()

    for scen in scens:
        atmosphere = scen["atmosphere"]

        atmosphere["o2o2"] = sk.constituent.CollisionInducedAbsorber(
            sk.optical.HITRANCollision("O2O2"), name="O2O2"
        )

        engine = sk.Engine(scen["config"], scen["geometry"], scen["viewing_geo"])

        radiance = sk.test_util.wf.numeric_wf(
            atmosphere.pressure_pa, 0.01, engine, atmosphere, "wf_pressure_pa"
        )

        sk.test_util.wf.validate_wf(
            radiance["wf_pressure_pa"],
            radiance["wf_pressure_pa_numeric"],
            wf_dim="altitude",
            decimal=4,
        )



from __future__ import annotations

import contextlib
from pathlib import Path

import numpy as np

import sasktran2 as sk
from sasktran2.constituent import VMRAltitudeAbsorber
from sasktran2.database.web import StandardDatabase
from sasktran2.optical.base import OpticalProperty


def _atm_file_path(folder_name: str, file_name: str) -> Path:
    """
    Determine the location of a specified '.atm' file.

    Parameters
    ----------
    folder_name : str
        Name of folder with the file, e.g. 'fascode', 'mipas_1998', or 'mipas_2001'. Do not include slashes
    file_name : str
        Name of the file, e.g. 'std.atm', 'tro.atm', ...

    Returns
    -------
    Path
        Absolute path to the file.
    """
    return StandardDatabase().path(
        (Path("climatology") / folder_name / file_name).as_posix()
    )


def _atm_reader(atm_file: str) -> dict:
    """
    Reads in a file with the extension '.atm' which contains information about atmospheric profiles and returns a
    dictionary with the atmospheric parameters as keys and the profiles as values. Refer to the input atm file for
    units of each profile but typically heights are in km, temperatures in K, pressures in mb, and species populations
    in ppmv.

    Parameters
    ----------
    atm_file : str

    Returns
    -------
    dict of [species: str, profile: np.array]
    """
    profiles = (
        {}
    )  # dictionary where keys are the species tags and values are the profiles
    with Path.open(atm_file) as f:
        num_levels_set = False
        cur_profile = ""
        for line in f:
            if line[0] == "!":  # line is a comment
                pass
            elif line[0] == "*":  # line is a profile header
                cur_profile = ""
                for letter in line[1::]:
                    if letter == " ":
                        break  # end of species name
                    cur_profile += letter
                if cur_profile == "END\n":
                    break  # end of file
                profiles[cur_profile.upper()] = np.array([])
            elif (
                not num_levels_set
            ):  # first uncommented line gives number of levels in each profile
                num_levels_set = True
            else:  # line contains profile values
                no_space = list(filter(None, line.split(" ")))
                no_space_or_comma = []
                for item in no_space:
                    no_space_or_comma.append(next(filter(None, item.split(","))))
                for item in no_space_or_comma:
                    with contextlib.suppress(ValueError):
                        if item != "\n":
                            profiles[cur_profile] = np.append(
                                profiles[cur_profile.upper()], float(item)
                            )
    return profiles


def constituent(
    species: str,
    optical_property: OpticalProperty,
    dataset: str = "fascode",
    climatology: str = "std",
) -> VMRAltitudeAbsorber:
    """
    Creates a VMRAltitudeAbsorber constituent for the given species and profile from several reference atmosphere models available at
    http://eodg.atm.ox.ac.uk/RFM/atm/

    Note: Species in the "Minor" category do not have a climatology associated VMR.  Instead they have single VMR profile

    The following table lists the species supported by each climatology.
    Each dataset contains several different climatologies. Each climatology contains unique profiles of temperature,
    pressure, and VMRs of the major species. The VMRs for the minor species are the same for all climatologies in each
    dataset.

    +------------+---------------------------------+---------------------------------+---------------------------------+
    | Dataset    | Climatologies                   | Major Species                   | Minor Species                   |
    +============+=================================+=================================+=================================+
    | fascode    | tro (Tropical),                 | H2O, CO2, O3, N2O, CO, CH4, O2  | NO, SO2, NO2, NH3, HNO3, OH,    |
    |            | mls (Mid-Latitude Summer),      |                                 | HF, HCl, HBr, HI, ClO, OCS,     |
    |            | mlw (Mid-Latitude Winter),      |                                 | H2CO, HOCl, N2, HCN, CH3Cl,     |
    |            | sas (Sub-Arctic Summer),        |                                 | H2O2, C2H2, C2H6, PH3, COF2,    |
    |            | saw (Sub-Arctic Winter),        |                                 | SF6, H2S, CFCl3, CF2Cl2, CClF3, |
    |            | std (US Standard Atmosphere),   |                                 | CF4, CHCl2F, CHClF2, C2Cl3F3,   |
    |            |                                 |                                 | C2Cl2F4, C2ClF5, CCl4, ClONO2,  |
    |            |                                 |                                 | N2O5, HNO4                      |
    +------------+---------------------------------+---------------------------------+---------------------------------+
    | mipas_1998 | day_imk (Mid-Latitude Day)      | N2, O2, O3P, CO2, O3, H2O, CH4, | HCN, H2O2, F12, F14, F22, COF2, |
    |            | ngt_imk (Mid-Latitude Night)    | N2O, HNO3, CO, NO2, N2O5, ClO,  | OCS, NH3, SO2, CFCl3, C2H2,     |
    |            | win_imk (Polar Winter)          | HOCl, ClONO2, NO                | C2H6, CCl4, SF6, HNO4, CH3Cl,   |
    |            | sum_imk (Polar Summer)          |                                 | CClF3, CHCl2F, C2Cl3F3, C2Cl2F4 |
    +------------+---------------------------------+---------------------------------+---------------------------------+
    | mipas_2001 | day (Mid-Latitude Day)          | N2, O2, CO2, O3, H2O, CH4, N2O, | CClF3, CHCl2F, C2Cl3F3,         |
    |            | ngt (Mid-Latitude Night)        | HNO3, CO, NO2, N2O5, ClO, HOCl, | C2Cl2F4, C2ClF5, CH3Cl, H2S     |
    |            | win (Polar Winter)              | ClONO2, NO, HNO4, HCN, NH3,     |                                 |
    |            | sum (Polar Summer)              | F11, F12, F14, F22, CCl4, COF2, |                                 |
    |            | equ (Equatorial Day)            | H2O2, C2H2, C2H6, OCS, SO2, SF6 |                                 |
    +------------+---------------------------------+---------------------------------+---------------------------------+

    Parameters
    ----------
    species : str
        The species to include, see the table below for supported species
    optical_property : OpticalProperty
        The optical property to use for the given species
    dataset : str, optional
        Dataset to use, see table below for options, by default "fascode"
    climatology : str, optional
        Climatology to use, see table below for options, by default "std"


    Returns
    -------
    sk.constituent.VMRAltitudeAbsorber
        The resulting constituent for the given species and profile
    """
    atm_file = climatology if climatology.endswith(".atm") else climatology + ".atm"
    file_path = _atm_file_path(dataset, atm_file)
    data = _atm_reader(file_path)

    if species.upper() in data:
        species_vmr = data[species.upper()] / 1.0e6  # Convert from ppm to vmr
        species_heights_m = data["HGT"] * 1000.0  # Convert from km to m
    else:
        # This is a minor species, need to load the minor species data
        if dataset == "mipas_2001":
            file_path = _atm_file_path(dataset, "extra.atm")
        elif dataset == "mipas_1998":
            file_path = _atm_file_path(dataset, "extra_imk.atm")
        elif dataset == "fascode":
            file_path = _atm_file_path(dataset, "minor.atm")
        else:
            msg = f"{dataset} is not a valid dataset"
            raise ValueError(msg)
        data_minor = _atm_reader(file_path)
        species_vmr = data_minor[species.upper()] / 1.0e6  # convert from ppm to vmr
        species_heights_m = data_minor["HGT"] * 1000.0  # convert from km to m

    return VMRAltitudeAbsorber(optical_property, species_heights_m, species_vmr)


def add_to_atmosphere(
    atmosphere: sk.Atmosphere,
    species: dict,
    dataset: str = "fascode",
    climatology: str = "std",
    set_pressure_temperature: bool = True,
):
    """
    Adds multiple species objects to an atmosphere object using the reference atmospheric profiles from http://eodg.atm.ox.ac.uk/RFM/atm/.

    See :py:func:`constituent` for more information on the species, dataset, and climatology options.

    Parameters
    ----------
    atmosphere : sk.Atmosphere
        The atmosphere to add the species objects to
    species : dict
        A dictionary where keys are Species (e.g. "O3") and values are the optical properties associated with them.  For example
        Species = {"O3": sk.optical.O3DBM(), "NO2": sk.optical.NO2Vandaele()}
    dataset : str, optional
        Reference dataset to use see :py:func:`constituent`, by default "fascode"
    climatology : str, optional
        Climatology to use, see :py:func:`constituent`, by default "std"
    set_pressure_temperature: bool, optional
        If true the atmospheric pressure/temperature will also be set by the climatology, by default True
    """
    for specie, optical_property in species.items():
        atmosphere[specie] = constituent(specie, optical_property, dataset, climatology)

    if set_pressure_temperature:
        atm_file = climatology if climatology.endswith(".atm") else climatology + ".atm"
        file_path = _atm_file_path(dataset, atm_file)
        data = _atm_reader(file_path)

        altitudes_m = data["HGT"] * 1000.0  # Convert from km to m
        temperature_k = data["TEM"]
        pressure_pa = data["PRE"] * 100.0

        atmosphere.temperature_k = np.interp(
            atmosphere.model_geometry.altitudes(), altitudes_m, temperature_k
        )
        # Interpolate pressure in log space
        atmosphere.pressure_pa = np.exp(
            np.interp(
                atmosphere.model_geometry.altitudes(), altitudes_m, np.log(pressure_pa)
            )
        )
